
//
// PHP4 language grammar for Sablecc3 parser generator.
// Copyright (C) 2001-2003 Indrek Mandre <indrek (at) mare . ee>
//
// This program is free software; you can redistribute it and/or modify it
// under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation; either version 2.1 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this program in the file "COPYING-LESSER"; if not,
// write to the Free Software Foundation, Inc., 59 Temple Place,
// Suite 330, Boston, MA  02111-1307  USA
//

package org.sablecc.php4;

import org.sablecc.php4.lexer.*;
import org.sablecc.php4.node.*;

import java.io.*;
import java.util.*;

public class PhpLexer extends Lexer
{
  private LinkedList list = new LinkedList ();
  private PushbackReader in;
  private boolean do_unescape_strings = true;

  /** Set whether unescape the PHP strings. By default is true. */
  public void setUnescapeStrings (boolean val)
  {
    do_unescape_strings = val;
  }

  public boolean getUnescapeStrings ()
  {
    return do_unescape_strings;
  }

  public PhpLexer (PushbackReader pbr)
  {
    super (pbr);
    in = pbr;

    if (false)
    while (true)
      {
        Token t = null;
        try {
          t = getToken();
        } catch (Exception e) {
          e.printStackTrace();
        }
        System.out.println (t.getClass().getName() + " [" + t.getText() + "]");
        if ( t instanceof EOF ) break;
      }
  }

  private boolean ifCheckColon (Collection add) throws LexerException, IOException
  {
    Token t = null;
    int pcount = 0;

    do {
        t = getToken();
        add.add (t);

        if ( t instanceof TLPar )
            pcount++;

        if ( t instanceof TRPar )
          {
            pcount--;
            if ( pcount == 0 )
                break;
          }

    } while ( !(t instanceof EOF || t instanceof TSemicolon) );

    if ( pcount != 0 )
        return false;

      // skip the blanks, find first nonblank and check it
    do {
        t = getToken ();
        add.add (t);
    } while (t instanceof TBlank);

    if ( t instanceof TColon )
        return true;

    return false;
  }

  private String current_heredoc_search;

  private Collection expand (Token tok) throws LexerException, IOException
  {
    LinkedList add = new LinkedList ();

    if ( tok instanceof TStringChunkHelper || tok instanceof THeredocChunkHelper || tok instanceof TBacktickChunkHelper )
      {
        tok = new TStaticString (tok.getText(), tok.getLine(), tok.getPos());
        add.add (tok);
        return add;
      }

    if ( tok instanceof THeredocStart )
      {
        // now now, weird system people hate me now because i take \n is eol symbol
        current_heredoc_search = "\n" + tok.getText ().substring(3) + ";";
        add.add (tok);
        return add;
      }

    if ( tok instanceof THeredocChunk )
      {
        String text = tok.getText();
        tok = new TStaticString (ds(text), tok.getLine(), tok.getPos());
        add.add (tok);
        int pos = text.indexOf (current_heredoc_search);
        if ( pos != -1 )
          {
            String read = text.substring (0, pos + current_heredoc_search.length());
            String unread = text.substring (pos + current_heredoc_search.length());
            int ladd = 0, index = -1;
            while ( (index = read.indexOf ('\n', index + 1)) != -1 ) ladd++;

            Token pushback = new TStaticString (text.substring (pos + current_heredoc_search.length() - 1),
                tok.getLine() + ladd, current_heredoc_search.length() + 1);
            unread (pushback);

            tok.setText (ds(text.substring (0, pos)));
            add.add (new THeredocEnd ("", tok.getLine(), tok.getPos()));

            state = State.CODE;
          }

        return add;
      }

    if ( tok instanceof TIf && ifCheckColon (add) )
      {
        tok = new TIfColon ("");
        add.add (0, tok);
        return add;
      }

    if ( token instanceof TElseif && ifCheckColon (add) )
      {
        tok = new TElseifColon ("");
        add.add (0, tok);
      }

    if ( tok instanceof TStaticString )
      {
        tok.setText (ss(tok.getText().substring(1, tok.getText().length() - 1)));
        add.add (tok);
        return add;
      }

    if ( tok instanceof TStringChunk || tok instanceof TBacktickChunk )
      {
        tok = new TStaticString (ds(tok.getText()), tok.getLine(), tok.getPos());
        add.add (tok);
        return add;
      }

    if ( tok instanceof TStringVar )
      {
        tok = new TVariable (tok.getText(), tok.getLine(), tok.getPos());
        add.add (tok);
        return add;
      }
    
    if ( tok instanceof TStringArrayStatic )
      {
        String str = tok.getText();
        String varname = str.substring (0, str.indexOf ('['));
        String keyvalue = str.substring (str.indexOf ('[') + 1, str.indexOf (']'));

        tok = new TStringCvarStart (tok.getText(), tok.getLine(), tok.getPos());
        add.add (tok);
        add.add (new TVariable (varname, tok.getLine(), tok.getPos()));
        add.add (new TLBracket (tok.getLine(), tok.getPos() + str.indexOf ('[')));
        add.add (new TStaticString (keyvalue, tok.getLine(), tok.getPos() + str.indexOf (']')));
        add.add (new TRBracket (tok.getLine(), tok.getPos() + str.indexOf (']')));
        add.add (new TStringCvarEnd ("", tok.getLine(), tok.getPos() + str.length()));
        return add;
      }

    if ( tok instanceof TStringArrayVar )
      {
        String str = tok.getText();
        String varname = str.substring (0, str.indexOf ('['));
        String keyvalue = str.substring (str.indexOf ('[') + 1, str.indexOf (']'));

        tok = new TStringCvarStart (tok.getText(), tok.getLine(), tok.getPos());
        add.add (tok);
        add.add (new TVariable (varname, tok.getLine(), tok.getPos()));
        add.add (new TLBracket (tok.getLine(), tok.getPos() + str.indexOf ('[')));
        add.add (new TVariable (keyvalue, tok.getLine(), tok.getPos() + str.indexOf (']')));
        add.add (new TRBracket (tok.getLine(), tok.getPos() + str.indexOf (']')));
        add.add (new TStringCvarEnd ("", tok.getLine(), tok.getPos() + str.length()));
        return add;
      }

    if ( tok instanceof TStringObject )
      {
        String str = tok.getText();
        String varname = str.substring (0, str.indexOf ('-'));
        String keyvalue = str.substring (str.indexOf ('>') + 1);

        tok = new TStringCvarStart (tok.getText(), tok.getLine(), tok.getPos());
        add.add (tok);
        add.add (new TVariable (varname, tok.getLine(), tok.getPos()));
        add.add (new TPointElem (tok.getLine(), tok.getPos() + str.indexOf ('-')));
        add.add (new TIdentifier (keyvalue, tok.getLine(), tok.getPos() + str.indexOf ('>') + 1));
        add.add (new TStringCvarEnd ("", tok.getLine(), tok.getPos() + str.length()));
        return add;
      }

    // thanks to this mess we can parse something like: echo "boogoo ${wof["yeehaa ${aa["here"]}"]}";
    if ( tok instanceof TStringToComplexCvar ||
        tok instanceof THeredocToComplexCvar ||
        tok instanceof TBacktickToComplexCvar )
      {
        String varname = tok.getText().substring (2);
        add.add (tok);
        add.add (new TLBrace (tok.getLine(), tok.getPos()));
        add.add (new TVariable ("$" + varname, tok.getLine(), tok.getPos()));

        int bc = 1;
        Token t;

        do {
            t = getToken();
            add.add (t);

            if ( t instanceof TLBrace )
                bc++;

            if ( t instanceof TRBrace )
              {
                bc--;
                if ( bc == 0 )
                    break;
              }
        } while ( !(t instanceof EOF) );

        if ( bc == 0 )
          {
            if ( tok instanceof TStringToComplexCvar )
                add.add (new TStringFromComplexCvar ("", t.getLine(), t.getPos() + 1));
            else if ( tok instanceof THeredocToComplexCvar )
                add.add (new THeredocFromComplexCvar ("", t.getLine(), t.getPos() + 1));
            else if ( tok instanceof TBacktickToComplexCvar )
                add.add (new TBacktickFromComplexCvar ("", t.getLine(), t.getPos() + 1));
            else
                throw new RuntimeException ("This should never happend!");
          }

        return add;
      }

    add.add (0, tok);
    return add;
  }

  protected Token getToken () throws IOException, LexerException
  {
    Token ret;

    if ( !list.isEmpty() )
      {
        ret = (Token)list.removeFirst();
      }
    else
      {
        ret = super.getToken();
        list.addAll (0, expand (ret));
        if ( list.isEmpty() ) return getToken();
        ret = (Token)list.removeFirst();
      }

    if ( ret instanceof TStringFromComplexCvar && list.isEmpty() )  // i wonder if this is really OK
        state = State.STRING;

    if ( ret instanceof THeredocFromComplexCvar && list.isEmpty() )  // i wonder if this is really OK
        state = State.HEREDOC;

    return ret;
  }

  /** Unescape single quote PHP string into a native java string. */
  public String ss (String s)
  {
    if ( !do_unescape_strings ) return s;

    StringBuffer str = new StringBuffer (s);
    int i = 0;
    while (i < str.length() - 1)
    {
      if ( str.charAt(i) == '\\' && str.charAt(i + 1) == '\'' )
        str.delete (i, i + 1);
      else if ( str.charAt(i) == '\\' && str.charAt(i + 1) == '\\' )
        str.delete (i, i + 1);
      i++;
    }

    return str.toString ();
  }

  /** Unescape double quote PHP string into a native java string. */
  public String ds (String s)
  {
    if ( !do_unescape_strings ) return s;

    StringBuffer str = new StringBuffer (s);
    int i = 0;
    while (i < str.length() - 1)
      {
        if ( str.charAt (i) == '\\' )
          {
            switch (str.charAt (i + 1))
              {
                case 'n':
                  str.replace (i, i + 2, "\n");
                  break;
                case 'r':
                  str.replace (i, i + 2, "\r");
                  break;
                case 't':
                  str.replace (i, i + 2, "\t");
                  break;
                case '\\':
                  str.replace (i, i + 2, "\\");
                  break;
                case '$':
                  str.replace (i, i + 2, "$");
                  break;
                case '"':
                  str.replace (i, i + 2, "\"");
                  break;
                default:
                  // check first for octal
                  char val = str.charAt (i + 1);
                  if ( val >= '0' && val <= '7' )
                    {
                      String octal = "0";
                      octal += val;

                      if ( i + 2 < str.length() )
                        {
                          val = str.charAt (i + 2);
                          if ( val >= '0' && val <= '7' )
                            {
                              octal += val;

                              if ( i + 3 < str.length() )
                                {
                                  val = str.charAt (i + 3);
                                  if ( val >= '0' && val <= '7' )
                                      octal += val;
                                }
                            }
                        }
                      try {
                        int v = Integer.decode (octal).intValue();
                        str.replace (i, i + octal.length(), "" + (char)v);
                      } catch (Exception e) { e.printStackTrace (); /* should never happen */ }
                    }
                  else if ( (val == 'x' || val == 'X') && i + 2 < str.length() )
                    {
                      val = Character.toLowerCase (str.charAt (i + 2));
                      if ( val >= 'a' && val <= 'f' || val >= '0' && val <= '9')
                        {
                          String hex = "0x";
                          hex += val;
                          if ( i + 3 < str.length() )
                            {
                              val = str.charAt (i + 3);
                              if ( val >= 'a' && val <= 'f' || val >= '0' && val <= '9')
                                  hex += val;
                            }

                          try {
                            int v = Integer.decode (hex).intValue ();
                            str.replace (i, i + hex.length(), "" + (char)v);
                          } catch (Exception e) { e.printStackTrace (); /* should never happen */ }
                        }
                    }
                  i++;
                  break;
              }
          }
        i++;
      }

    return str.toString ();
  }
};

